// Aggregate.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: July 2009

package eu.reservoir.aggregator;

import java.util.Collection;
import java.util.HashSet;
import java.util.TreeSet;


/**
 * Function	Return Type	
 * sum	Number	
 * average	Number	
 * length	Number	
 * max(n)	Stream	max(5) returns 5 highest values
 * min(n)	Stream	min(1) returns lowest value
 * sd	Number	standard deviation
 * test(fn, value, %age)	Boolean	test(<, 30, 95) checks if 95% of values < 30
 */

public class Aggregate {
    /**
     * Get the sum of a collection of numbers.
     */
    public Number sum(Collection<Number> collection) {
	Double sum = 0.0;

	for (Number n : collection) {
	    sum += n.doubleValue();
	}

	return sum;
    }

    /**
     * Get the length of a collection of numbers.
     */
    public Number length(Collection<Number> collection) {
	Number n = collection.size();

	return n;
	
    }

    /**
     * Get the average of a collection of numbers.
     */
    public Number average(Collection<Number> collection) {
	Number sum = sum(collection);
	Number n = collection.size();

	return sum.doubleValue() / n.doubleValue();
	
    }

    /**
     * Get the variance of a collection of numbers.
     *
     * The variance (σ2) is a measure of how far each value in the
     * data set is from the mean. Here is how it is defined:
     * 
     * Determine sum squares.
     * Divide by the number of elememnts =>  sum(x * x) / N
     * Subtract mu squared.
     * The standard deviation (σ) is simply the (positive) 
     * square root of the variance.
     */
    public Number variance(Collection<Number> collection) {
	double sumSq = 0.0;
	double mu = average(collection).doubleValue();
	double muSq = mu * mu;
	int n = length(collection).intValue();

	for (Number x : collection) {
	    sumSq += (x.doubleValue() * x.doubleValue());
	}

	return (sumSq / n) - muSq;
    }

    /**
     * Get the standard deviation of a collection of numbers.
     */
    public Number sd(Collection<Number> collection) {
	return Math.sqrt(variance(collection).doubleValue());
    }

    /**
     * Find the N highest numbers in the collection of numbers.
     */
    public Collection<Number> max(int n, Collection<Number> collection) {
	TreeSet<Number> maxN = new TreeSet<Number>();

	for (Number x : collection) {
	    // check if maxN has reached size N
	    if (maxN.size() == n) {
		// if it has we need to see if x should go in it
		// maxN is ordered, so if x >= first element of maxN
		// then x needs to go in maxN
		Number first = maxN.first();
		if (x.doubleValue() >= first.doubleValue()) {
		    // remove first
		    maxN.remove(first);
		    // add x
		    maxN.add(x);
		} else {
		    // nothing to do
		}

	    } else { 
		// if it is not size N, always add x
		maxN.add(x);
	    }
	}

	return maxN;
    }


    /**
     * Find the N lowest numbers in the collection of numbers.
     */
    public Collection<Number> min(int n, Collection<Number> collection) {
	TreeSet<Number> minN = new TreeSet<Number>();

	for (Number x : collection) {
	    // check if minN has reached size N
	    if (minN.size() == n) {
		// if it has we need to see if x should go in it
		// minN is ordered, so if x < last element of minN
		// then x needs to go in minN
		Number last = minN.last();
		if (x.doubleValue() < last.doubleValue()) {
		    // remove first
		    minN.remove(last);
		    // add x
		    minN.add(x);
		} else {
		    // nothing to do
		}

	    } else { 
		// if it is not size N, always add x
		minN.add(x);
	    }
	}

	return minN;
    }

    /**
     * Find the highest number in the collection of numbers.
     */
    public Number max(Collection<Number> collection) {
	return max(1, collection).iterator().next();
    }

    /**
     * Find the lowest number in the collection of numbers.
     */
    public Number min(Collection<Number> collection) {
	return min(1, collection).iterator().next();
    }

    /**
     * Test a value
     */
    public boolean test(Function1 fn, Number n) {
	return fn.call(n.doubleValue());
    }

    /**
     * Filter the ones that match the predicate.
     */
    public Collection<Number> filter(Function1 predicate, Collection<Number> collection) {
	HashSet<Number> result = new HashSet<Number>();

	for (Number x : collection) {
	    if (predicate.call(x)) {
		// add to result
		result.add(x);
	    }
	}

	return result;
    }


    /**
     * Test if a proportion of the list tests true wrt the nominate value.
     * Proportion is in range 0 to 1
     */
    public boolean testProportion(Function2 function, Number value, double proportion, Collection<Number> collection) {
	if (proportion < 0 || proportion > 1) {
	    throw new RuntimeException("Proportion is " + proportion + ".  Allowable range is 0 -> 1");
	}

	int count = 0;

	// count the number < value
	for (Number x : collection) {
	    if (function.call(x, value)) {
		count++;
	    }
	}

	// the calculated proportion is count / length(collection)
	Number length = length(collection);
	double calculatedProportion = count / length.doubleValue();

	//System.out.println("count = " + count + " length = " + length + " calc_proportion = " + calculatedProportion);

	if (calculatedProportion >= proportion) {
	    return true;
	} else {
	    return false;
	}

    }


    public static void main(String[] args) {
	HashSet<Number> set = new HashSet<Number>();

	set.add(11.1);
	set.add(12.2);
	set.add(13.3);
	set.add(14.4);
	set.add(15.5);
	set.add(16.6);
	set.add(17.7);
	set.add(18.8);
	set.add(19.9);

	aggregate(set);
    }

    public static void aggregate(Collection<Number> set) {
	Aggregate agg = new Aggregate();

	System.out.println("set = " + set);
	System.out.printf("sum = %8.3f\n", agg.sum(set));
	System.out.printf("length = %d\n", agg.length(set));
	System.out.printf("average = %8.3f\n", agg.average(set));
	System.out.printf("variance = %8.3f\n", agg.variance(set));
	System.out.printf("sd = %8.3f\n", agg.sd(set));
	System.out.println("max(3) = " + agg.max(3, set));
	System.out.println("max = " + agg.max(set));

	System.out.println("min(3) = " + agg.min(3, set));
	System.out.println("min = " + agg.min(set));

	Function1 lt15 = new Function1() {
		public boolean call(Number n1) {
		    return n1.doubleValue() < 15;
		}
	    };

	System.out.println("filter < 15 = " + agg.filter(lt15, set));

	Function2 lt = new Function2() {
		public boolean call(Number n1, Number n2) {
		    return n1.doubleValue() < n2.doubleValue();
		}
	    };
	Function2 gte = new Function2() {
		public boolean call(Number n1, Number n2) {
		    return n1.doubleValue() >= n2.doubleValue();
		}
	    };
	System.out.println("testProportion(<, 15, 50%) = " + agg.testProportion(lt, 15, 0.50, set));
	System.out.println("testProportion(>=, 15, 50%) = " + agg.testProportion(gte, 15, 0.50, set));

	// average of the 3 highest values > 15
	System.out.println("average(max(3, set)) = " + agg.average(agg.max(3, set)));
	System.out.println("average(max(3, set)) < 15 = " + agg.test(lt15, (agg.average(agg.max(3, set)))));
    }


    /**
     * A function that takes one arg.
     */
    public interface Function1 {
	public boolean call(Number n1);
    }

    /**
     * A function that takes two args.
     */
    public interface Function2 {
	public boolean call(Number n1, Number n2);
    }
}

