// AggregationPoint.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: July 2009

package eu.reservoir.aggregator;

import eu.reservoir.monitoring.core.*;
import eu.reservoir.monitoring.core.plane.DataPlane;
import eu.reservoir.monitoring.distribution.*;
//import eu.reservoir.monitoring.im.*;
import eu.reservoir.monitoring.appl.datarate.*;
import eu.reservoir.monitoring.appl.DynamicControl;
import com.timeindexing.time.*;
import com.timeindexing.basic.*;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.io.Serializable;

/**
 * An AggregationPoint.
 * This collects measurement data from one data distribution plane,
 * at a regular interval it selects some of the measurements,
 * it aggregates the selected data, and then forwards the result
 * onto a different data distribution plane.
 */
public class AggregationPoint extends DynamicControl {
    // the collector of the measurements
    Collector collector;
    
    // the selector of measurements
    Selector selector;

    // the actual aggregator of the selected measurements
    Aggregator aggregator;

    // the forwarder of the aggregated result
    Forwarder forwarder;

    // when the AggregationPoint should wake up
    Timing when;

    // what should the AggregationPoint select in order to aggregate
    Interval what;

    /**
     * Construct an AggregationPoint.
     */
    public AggregationPoint(DataPlane inDataPlane,
			    DataPlane outDataPlane,
			    String name, String collectorPath,
			    Extractor extractor, Chooser chooser,
			    Timing when,
			    Interval what,
			    AggregateFn aggFn,
			    List<ProbeAttribute> probeAttributes
			    ) {
	super(name);

	// set up Collector
	collector = new Collector(inDataPlane, name, collectorPath);
	// set the extractor
	collector.setExtractor(extractor);

	// set up Selector
	selector = new Selector(collector.getIndex());
	// set the chooser
	selector.setChooser(chooser);

	// set up an Aggregator
	aggregator = new Aggregator();
	// set the aggregation function
	aggregator.setAggregateFn(aggFn);

	// set up a Forwarder
	forwarder = new Forwarder(outDataPlane, "aggregated", name, collectorPath, probeAttributes);

	this.when = when;
	this.what = what;

	// set sleepTime in seconds
	if (when instanceof Timing.Regular) {
	    // set a regular timeout
	    setSleepTime(((Timing.Regular)when).getSeconds());
	} else {
	    // NB. there is no code yet for Timing.Specific
	    throw new Error("You can't set specific timings yet!");
	}

    }

    /**
     * Construct an AggregationPoint.
     */
    public AggregationPoint(DataPlane inDataPlane,
			    DataPlane outDataPlane,
			    String name, String collectorPath,
			    Extractor extractor, Chooser chooser,
			    Filter filter,
			    Timing when,
			    Interval what,
			    AggregateFn aggFn,
			    List<ProbeAttribute> probeAttributes
			    ) {

	this(inDataPlane, outDataPlane, name, collectorPath, extractor, chooser, when, what, aggFn, probeAttributes);

	forwarder.setFilter(filter);
    }


    /**
     * Initialize.
     */
    protected void controlInitialize() {
	collector.start();
	forwarder.start();

	// the sleep time is done after the controlEvaluate() function
	// in the main loop, but we need to explicitly do one,
	// so do a sleep first
	try {
	    //System.err.println("Entering sleep " + timeToSleep + " @ " + System.currentTimeMillis());
	    sleep();
	    //System.err.println("Exiting sleep " + timeToSleep + " @ " + System.currentTimeMillis());
	} catch (InterruptedException ie) {
	    //System.err.println(getName() + ": sleep interrupted");
	} 

    }

    /**
     * Actually evaluate something.
     */
    protected void controlEvaluate() {
	Collection<ChooserResult> numbers = selector.select(what);

	if (numbers.size() == 0) {
	    // there is nothing to do
	} else {
	    // aggregate and forward
	    AggregatorMeasurement result = aggregator.aggregate(numbers);
	    forwarder.setValue(result);
	}
    }

    /**
     * Cleanup
     */
    protected void controlCleanup() {
	collector.stop();
	forwarder.stop();
    }


}
