// AP1.java
// Author: Stuart Clayman
// Email: sclayman@ee.ucl.ac.uk
// Date: July 2009

package eu.reservoir.aggregator.tests;

import eu.reservoir.aggregator.*;
import eu.reservoir.monitoring.core.*;
import eu.reservoir.monitoring.core.plane.DataPlane;
import eu.reservoir.monitoring.distribution.*;
import eu.reservoir.monitoring.distribution.multicast.*;
//import eu.reservoir.monitoring.im.*;
import eu.reservoir.monitoring.appl.DynamicControl;
import com.timeindexing.time.*;
import com.timeindexing.basic.*;
import java.util.Collection;
import java.util.ArrayList;
import java.io.Serializable;


/**
 * Test of AggregationPoint.
 * This aggregates measurements by averaging data for the last 30 seconds,
 * every 30 seconds.
 * It listens on MulticastAddress("229.229.0.1", 2299) and forwards
 * to  MulticastAddress("229.229.0.1", 2288)
 */
public class AP1 {

    /**
     * Main entry point
     */
    public static void main(String[] args) {
	String name = "agg-point1";

	/*
	 * Set up the Collector
	 */

	// The multicast address for the collector.
	MulticastAddress inputDataAddress = new MulticastAddress("229.229.0.1", 2299);
        DataPlane inputDataPlane = new MulticastDataPlaneConsumer(inputDataAddress);

	// The place to store the collected data
	String collectorPath = "/tmp/";

	// The extraction function, which takes an individual Measurements
	// and returns the data to be stored by the Collector
	// In this case it is the value returned by Measurement::getValues()
	// which is an ArrayList<ProbeValue> 
	Extractor onlyProbeValues = new Extractor() {
		public Serializable extract(Measurement m) {
		    Serializable object = (Serializable)m.getValues();
		    System.out.println("Collected: " + object);
		    return object;
		}
	    };


	/*
	 * Set up the Selector
	 */
	// The chooser function is applied by the Selector to values
	// collected from the Extractor.
	// In this case it takes an ArrayList<ProbeValue> and
	// returns value of first one, which is a Number
	Chooser firstValue = new Chooser() {
		public ChooserResult choose(Object obj) {
		    ArrayList<ProbeValue> list = (ArrayList<ProbeValue>)obj;
		    return new NumberCR((Number)list.get(0).getValue());
		}
	    };

	/*
	 * Set up the Aggregator
	 */

	// This takes the Collection<Number> that is returned by
	// the Selector and aggregates them.
	// In this case it is the average function.
	AggregateFn average = new AggregateFn() {
		public AggregatorMeasurement aggregate(Collection<ChooserResult> coll) {
		    // convert a Collection<ChooserResult> into Collection<Number>
		    ArrayList list = new ArrayList();
		    for (ChooserResult cr : coll) {
			list.add(((NumberCR)cr).number);
		    }

		    // aggregate the numbers
		    Aggregate aggregate = new Aggregate();
		    // to get the average
		    Number avg =  aggregate.average(list);
		    System.out.println("Aggregation: average = " + avg);

		    // now create a Measurement as the result

		    try {
			// field 0 is the avg
			DefaultProbeValue pv0 = new DefaultProbeValue(0, avg);
		    
			// create a list of ProbeValue
			ArrayList pvList = new ArrayList();
			pvList.add(pv0);

			// Construct the Measurement
			AggregatorMeasurement m = new AggregatorMeasurement(new eu.reservoir.monitoring.core.ID(1456), "aggregated-average", pvList);

			// return it
			return m;
		    } catch (Exception e) {
			return null;
		    }
		    
		}
	    };



	/*
	 * Set up the Forwarder
	 */

	// The multicast address for the forwarder
	MulticastAddress outputDataAddress = new MulticastAddress("229.229.0.1", 2288);
        DataPlane outputDataPlane = new MulticastDataPlaneProducer(outputDataAddress);


	// set up ProbeAttributes
	// field 0 is the avg
	DefaultProbeAttribute pa0 = new DefaultProbeAttribute(0, "average", ProbeAttributeType.DOUBLE, "n");
		    
	// create a list of ProbeAttribute
	ArrayList paList = new ArrayList();
	paList.add(pa0);




	/*
	 * Set up aggregation
	 */

	// Here the AggregationPoint will wake up every 30 seconds
	// and select all the data from the last measurement,
	// back 30 seconds.

	// Set sleep time to be 30 seconds
	Timing.Regular sleepTime = new SleepSeconds(30);

	// The selection of data to get, 
	// Get last 30 seconds of collected data, as per sleepTime
	Second backN = new Second(sleepTime.getSeconds(), TimeDirection.BACKWARD);
	// From Position.END_OF_INDEX, back 30 seconds
	Interval interval = new EndPointInterval((AbsolutePosition)Position.END_OF_INDEX, backN);


	// Allocate the AggregationPoint
	AggregationPoint aggPoint = new AggregationPoint(inputDataPlane,
							 outputDataPlane,
							 name,
							 collectorPath,
							 onlyProbeValues,
							 firstValue,
							 sleepTime,
							 interval,
							 average,
							 paList);

	// and activate it.
	aggPoint.activateControl();
    }

    // A ChooserResult class.
    // This one is a wrapper for a Number.
    static class NumberCR implements ChooserResult {
	public Number number;

	public NumberCR(Number n) {
	    number = n;
	}
    }

}
